"""Implementation of KNX 1-Bit KNX values."""

from __future__ import annotations

from .dpt import DPTEnum, DPTEnumData, EnumDataT
from .payload import DPTBinary


class DPT1BitEnum(DPTEnum[EnumDataT]):
    """Base class for KNX 1-Bit values encoded as Enums."""

    payload_type = DPTBinary
    payload_length = 1
    dpt_main_number = 1


class Switch(DPTEnumData):
    """Enum for switching control."""

    OFF = False
    ON = True


class DPTSwitch(DPT1BitEnum[Switch]):
    """Abstraction for KNX 1-Bit switch value."""

    dpt_main_number = 1
    dpt_sub_number = 1
    value_type = "switch"
    data_type = Switch

    @classmethod
    def _to_knx(cls, value: Switch) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Bool(DPTEnumData):
    """Enum for switching control."""

    FALSE = False
    TRUE = True


class DPTBool(DPT1BitEnum[Bool]):
    """Abstraction for KNX 1-Bit bool value."""

    dpt_main_number = 1
    dpt_sub_number = 2
    value_type = "bool"
    data_type = Bool

    @classmethod
    def _to_knx(cls, value: Bool) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Enable(DPTEnumData):
    """Enum for enable control."""

    DISABLE = False
    ENABLE = True


class DPTEnable(DPT1BitEnum[Enable]):
    """Abstraction for KNX 1-Bit enable value."""

    dpt_main_number = 1
    dpt_sub_number = 3
    value_type = "enable"
    data_type = Enable

    @classmethod
    def _to_knx(cls, value: Enable) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Ramp(DPTEnumData):
    """Enum for ramp control."""

    NO_RAMP = False
    RAMP = True


class DPTRamp(DPT1BitEnum[Ramp]):
    """Abstraction for KNX 1-Bit ramp value."""

    dpt_main_number = 1
    dpt_sub_number = 4
    value_type = "ramp"
    data_type = Ramp

    @classmethod
    def _to_knx(cls, value: Ramp) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Alarm(DPTEnumData):
    """Enum for alarm control."""

    NO_ALARM = False
    ALARM = True


class DPTAlarm(DPT1BitEnum[Alarm]):
    """Abstraction for KNX 1-Bit alarm value."""

    dpt_main_number = 1
    dpt_sub_number = 5
    value_type = "alarm"
    data_type = Alarm

    @classmethod
    def _to_knx(cls, value: Alarm) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class BinaryValue(DPTEnumData):
    """Enum for binary value control."""

    LOW = False
    HIGH = True


class DPTBinaryValue(DPT1BitEnum[BinaryValue]):
    """Abstraction for KNX 1-Bit binary_value value."""

    dpt_main_number = 1
    dpt_sub_number = 6
    value_type = "binary_value"
    data_type = BinaryValue

    @classmethod
    def _to_knx(cls, value: BinaryValue) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Step(DPTEnumData):
    """Enum for dimming control."""

    DECREASE = False
    INCREASE = True


class DPTStep(DPT1BitEnum[Step]):
    """Abstraction for KNX 1-Bit step value."""

    dpt_main_number = 1
    dpt_sub_number = 7
    value_type = "step"
    data_type = Step

    @classmethod
    def _to_knx(cls, value: Step) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class UpDown(DPTEnumData):
    """Enum for up/down."""

    UP = False
    DOWN = True


class DPTUpDown(DPT1BitEnum[UpDown]):
    """Abstraction for KNX 1-Bit up/down value."""

    dpt_main_number = 1
    dpt_sub_number = 8
    value_type = "up_down"
    data_type = UpDown

    @classmethod
    def _to_knx(cls, value: UpDown) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class OpenClose(DPTEnumData):
    """Enum for dimming control."""

    OPEN = False
    CLOSE = True


class DPTOpenClose(DPT1BitEnum[OpenClose]):
    """Abstraction for KNX 1-Bit open_close value."""

    dpt_main_number = 1
    dpt_sub_number = 9
    value_type = "open_close"
    data_type = OpenClose

    @classmethod
    def _to_knx(cls, value: OpenClose) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Start(DPTEnumData):
    """Enum for dimming control."""

    STOP = False
    START = True


class DPTStart(DPT1BitEnum[Start]):
    """Abstraction for KNX 1-Bit start value."""

    dpt_main_number = 1
    dpt_sub_number = 10
    value_type = "start"
    data_type = Start

    @classmethod
    def _to_knx(cls, value: Start) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class State(DPTEnumData):
    """Enum for dimming control."""

    INACTIVE = False
    ACTIVE = True


class DPTState(DPT1BitEnum[State]):
    """Abstraction for KNX 1-Bit state value."""

    dpt_main_number = 1
    dpt_sub_number = 11
    value_type = "state"
    data_type = State

    @classmethod
    def _to_knx(cls, value: State) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Invert(DPTEnumData):
    """Enum for dimming control."""

    NOT_INVERTED = False
    INVERTED = True


class DPTInvert(DPT1BitEnum[Invert]):
    """Abstraction for KNX 1-Bit invert value."""

    dpt_main_number = 1
    dpt_sub_number = 12
    value_type = "invert"
    data_type = Invert

    @classmethod
    def _to_knx(cls, value: Invert) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class DimSendStyle(DPTEnumData):
    """Enum for dimming control."""

    START_STOP = False
    CYCLICALLY = True


class DPTDimSendStyle(DPT1BitEnum[DimSendStyle]):
    """Abstraction for KNX 1-Bit dim_send_style value."""

    dpt_main_number = 1
    dpt_sub_number = 13
    value_type = "dim_send_style"
    data_type = DimSendStyle

    @classmethod
    def _to_knx(cls, value: DimSendStyle) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class InputSource(DPTEnumData):
    """Enum for dimming control."""

    FIXED = False
    CALCULATED = True


class DPTInputSource(DPT1BitEnum[InputSource]):
    """Abstraction for KNX 1-Bit input_source value."""

    dpt_main_number = 1
    dpt_sub_number = 14
    value_type = "input_source"
    data_type = InputSource

    @classmethod
    def _to_knx(cls, value: InputSource) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Reset(DPTEnumData):
    """Enum for dimming control."""

    NO_ACTION = False
    RESET = True


class DPTReset(DPT1BitEnum[Reset]):
    """Abstraction for KNX 1-Bit reset value."""

    dpt_main_number = 1
    dpt_sub_number = 15
    value_type = "reset"
    data_type = Reset

    @classmethod
    def _to_knx(cls, value: Reset) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Ack(DPTEnumData):
    """Enum for dimming control."""

    NO_ACTION = False
    ACKNOWLEDGE = True


class DPTAck(DPT1BitEnum[Ack]):
    """Abstraction for KNX 1-Bit ack value."""

    dpt_main_number = 1
    dpt_sub_number = 16
    value_type = "ack"
    data_type = Ack

    @classmethod
    def _to_knx(cls, value: Ack) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Trigger(DPTEnumData):
    """Enum for dimming control."""

    TRIGGER_0 = False
    TRIGGER = True


class DPTTrigger(DPT1BitEnum[Trigger]):
    """Abstraction for KNX 1-Bit trigger value."""

    dpt_main_number = 1
    dpt_sub_number = 17
    value_type = "trigger"
    data_type = Trigger

    @classmethod
    def _to_knx(cls, value: Trigger) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class Occupancy(DPTEnumData):
    """Enum for dimming control."""

    NOT_OCCUPIED = False
    OCCUPIED = True


class DPTOccupancy(DPT1BitEnum[Occupancy]):
    """Abstraction for KNX 1-Bit occupancy value."""

    dpt_main_number = 1
    dpt_sub_number = 18
    value_type = "occupancy"
    data_type = Occupancy

    @classmethod
    def _to_knx(cls, value: Occupancy) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class WindowDoor(DPTEnumData):
    """Enum for dimming control."""

    CLOSED = False
    OPEN = True


class DPTWindowDoor(DPT1BitEnum[WindowDoor]):
    """Abstraction for KNX 1-Bit window_door value."""

    dpt_main_number = 1
    dpt_sub_number = 19
    value_type = "window_door"
    data_type = WindowDoor

    @classmethod
    def _to_knx(cls, value: WindowDoor) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class LogicalFunction(DPTEnumData):
    """Enum for dimming control."""

    OR = False
    AND = True


class DPTLogicalFunction(DPT1BitEnum[LogicalFunction]):
    """Abstraction for KNX 1-Bit logical_function value."""

    dpt_main_number = 1
    dpt_sub_number = 21
    value_type = "logical_function"
    data_type = LogicalFunction

    @classmethod
    def _to_knx(cls, value: LogicalFunction) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class SceneAB(DPTEnumData):
    """Enum for dimming control."""

    SCENE_A = False
    SCENE_B = True


class DPTSceneAB(DPT1BitEnum[SceneAB]):
    """Abstraction for KNX 1-Bit scene_ab value."""

    dpt_main_number = 1
    dpt_sub_number = 22
    value_type = "scene_ab"
    data_type = SceneAB

    @classmethod
    def _to_knx(cls, value: SceneAB) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class ShutterBlindsMode(DPTEnumData):
    """Enum for dimming control."""

    UP_DOWN_MODE = False
    STEP_STOP_MODE = True


class DPTShutterBlindsMode(DPT1BitEnum[ShutterBlindsMode]):
    """Abstraction for KNX 1-Bit shutter_blinds_mode value."""

    dpt_main_number = 1
    dpt_sub_number = 23
    value_type = "shutter_blinds_mode"
    data_type = ShutterBlindsMode

    @classmethod
    def _to_knx(cls, value: ShutterBlindsMode) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class DayNight(DPTEnumData):
    """Enum for dimming control."""

    DAY = False
    NIGHT = True


class DPTDayNight(DPT1BitEnum[DayNight]):
    """Abstraction for KNX 1-Bit day_night value."""

    dpt_main_number = 1
    dpt_sub_number = 24
    value_type = "day_night"
    data_type = DayNight

    @classmethod
    def _to_knx(cls, value: DayNight) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class HeatCool(DPTEnumData):
    """Enum for heat/cool."""

    COOL = False
    HEAT = True


class DPTHeatCool(DPT1BitEnum[HeatCool]):
    """Abstraction for KNX 1-Bit heat/cool value."""

    dpt_main_number = 1
    dpt_sub_number = 100
    value_type = "heat_cool"
    data_type = HeatCool

    @classmethod
    def _to_knx(cls, value: HeatCool) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class ConsumerProducer(DPTEnumData):
    """Enum for dimming control."""

    CONSUMER = False
    PRODUCER = True


class DPTConsumerProducer(DPT1BitEnum[ConsumerProducer]):
    """Abstraction for KNX 1-Bit consumer_producer value."""

    dpt_main_number = 1
    dpt_sub_number = 1200
    value_type = "consumer_producer"
    data_type = ConsumerProducer

    @classmethod
    def _to_knx(cls, value: ConsumerProducer) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)


class EnergyDirection(DPTEnumData):
    """Enum for dimming control."""

    POSITIVE = False
    NEGATIVE = True


class DPTEnergyDirection(DPT1BitEnum[EnergyDirection]):
    """Abstraction for KNX 1-Bit energy_direction value."""

    dpt_main_number = 1
    dpt_sub_number = 1201
    value_type = "energy_direction"
    data_type = EnergyDirection

    @classmethod
    def _to_knx(cls, value: EnergyDirection) -> DPTBinary:
        """Return the raw KNX value for an Enum member."""
        return DPTBinary(value.value)



"""
Module for serialization and deserialization of APCI payloads.

APCI stands for Application Layer Protocol Control Information.

An APCI payload contains a service and payload. For example, a GroupValueWrite
is a service that takes a DPT as a value.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
import struct
from typing import ClassVar, cast

from xknx.dpt import DPTArray, DPTBinary
from xknx.exceptions import ConversionError
from xknx.secure.data_secure_asdu import SecureData, SecurityControlField
from xknx.telegram.address import IndividualAddress


def encode_cmd_and_payload(
    cmd: APCIService | APCIUserService | APCIExtendedService,
    encoded_payload: int = 0,
    appended_payload: bytes | None = None,
) -> bytearray:
    """Encode cmd and payload."""
    data = bytearray(
        [
            (cmd.value >> 8) & 0b11,
            (cmd.value & 0xFF) | (encoded_payload & DPTBinary.APCI_BITMASK),
        ]
    )
    if appended_payload:
        data.extend(appended_payload)
    return data


class APCIService(Enum):
    """Enum class for APCI services."""

    GROUP_READ = 0x0000
    GROUP_RESPONSE = 0x0040
    GROUP_WRITE = 0x0080

    INDIVIDUAL_ADDRESS_WRITE = 0x00C0
    INDIVIDUAL_ADDRESS_READ = 0x0100
    INDIVIDUAL_ADDRESS_RESPONSE = 0x140

    ADC_READ = 0x0180
    ADC_RESPONSE = 0x1C0

    MEMORY_EXTENDED_WRITE = 0x1FB
    MEMORY_EXTENDED_WRITE_RESPONSE = 0x1FC
    MEMORY_EXTENDED_READ = 0x1FD
    MEMORY_EXTENDED_READ_RESPONSE = 0x1FE

    MEMORY_READ = 0x0200
    MEMORY_RESPONSE = 0x0240
    MEMORY_WRITE = 0x0280

    USER_MESSAGE = 0x02C0

    DEVICE_DESCRIPTOR_READ = 0x0300
    DEVICE_DESCRIPTOR_RESPONSE = 0x0340

    RESTART = 0x0380

    ESCAPE = 0x03C0


class APCIUserService(Enum):
    """Enum class for user message APCI services."""

    USER_MEMORY_READ = 0x02C0
    USER_MEMORY_RESPONSE = 0x02C1
    USER_MEMORY_WRITE = 0x02C2

    USER_MANUFACTURER_INFO_READ = 0x02C5
    USER_MANUFACTURER_INFO_RESPONSE = 0x02C6

    FUNCTION_PROPERTY_COMMAND = 0x02C7
    FUNCTION_PROPERTY_STATE_READ = 0x02C8
    FUNCTION_PROPERTY_STATE_RESPONSE = 0x02C9


class APCIExtendedService(Enum):
    """Enum class for extended APCI services."""

    AUTHORIZE_REQUEST = 0x03D1
    AUTHORIZE_RESPONSE = 0x03D2

    PROPERTY_VALUE_READ = 0x03D5
    PROPERTY_VALUE_RESPONSE = 0x03D6
    PROPERTY_VALUE_WRITE = 0x03D7

    PROPERTY_DESCRIPTION_READ = 0x03D8
    PROPERTY_DESCRIPTION_RESPONSE = 0x03D9

    INDIVIDUAL_ADDRESS_SERIAL_READ = 0x03DC
    INDIVIDUAL_ADDRESS_SERIAL_RESPONSE = 0x03DD
    INDIVIDUAL_ADDRESS_SERIAL_WRITE = 0x03DE

    # DataSecure
    APCI_SEC = 0x03F1


@dataclass(slots=True)
class APCI(ABC):
    """
    Base class for ACPI services.

    This base class is only the interface for the derived classes.
    """

    CODE: ClassVar[APCIService | APCIUserService | APCIExtendedService] = cast(
        APCIService, None
    )

    @abstractmethod
    def calculated_length(self) -> int:
        """Get length of APCI payload - to be implemented in derived class."""

    @abstractmethod
    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data - to be implemented in derived class."""
        # shall return bytearray instead of bytes so TPCI can be
        # added to first 6 bits of first byte in-place later

    @classmethod
    @abstractmethod
    def from_knx(cls, raw: bytes) -> APCI:
        """
        Parse/deserialize from KNX/IP raw data - to be implemented in derived class.

        `raw` shall be a complete APDU.
        Return APCI instance based on APCI service.

        There are only 16 possible APCI services. The
        `APCIService.USER_MESSAGE` and `APCIService.ESCAPE` service have
        several sub-services.
        """
        apci = (raw[0] * 256 + raw[1]) & 0x03FF
        service = apci & 0x03C0

        if service == APCIService.GROUP_READ.value:
            return GroupValueRead.from_knx(raw)
        if service == APCIService.GROUP_WRITE.value:
            return GroupValueWrite.from_knx(raw)
        if service == APCIService.GROUP_RESPONSE.value:
            return GroupValueResponse.from_knx(raw)
        if service == APCIService.INDIVIDUAL_ADDRESS_WRITE.value:
            return IndividualAddressWrite.from_knx(raw)
        if service == APCIService.INDIVIDUAL_ADDRESS_READ.value:
            return IndividualAddressRead.from_knx(raw)
        if service == APCIService.INDIVIDUAL_ADDRESS_RESPONSE.value:
            return IndividualAddressResponse.from_knx(raw)
        if service == APCIService.ADC_READ.value:
            return ADCRead.from_knx(raw)
        if service == APCIService.ADC_RESPONSE.value:
            if apci == APCIService.MEMORY_EXTENDED_WRITE.value:
                return MemoryExtendedWrite.from_knx(raw)
            if apci == APCIService.MEMORY_EXTENDED_WRITE_RESPONSE.value:
                return MemoryExtendedWriteResponse.from_knx(raw)
            if apci == APCIService.MEMORY_EXTENDED_READ.value:
                return MemoryExtendedRead.from_knx(raw)
            if apci == APCIService.MEMORY_EXTENDED_READ_RESPONSE.value:
                return MemoryExtendedReadResponse.from_knx(raw)
            return ADCResponse.from_knx(raw)
        if service == APCIService.MEMORY_READ.value:
            return MemoryRead.from_knx(raw)
        if service == APCIService.MEMORY_WRITE.value:
            return MemoryWrite.from_knx(raw)
        if service == APCIService.MEMORY_RESPONSE.value:
            return MemoryResponse.from_knx(raw)
        if service == APCIService.USER_MESSAGE.value:
            if apci == APCIUserService.USER_MEMORY_READ.value:
                return UserMemoryRead.from_knx(raw)
            if apci == APCIUserService.USER_MEMORY_RESPONSE.value:
                return UserMemoryResponse.from_knx(raw)
            if apci == APCIUserService.USER_MEMORY_WRITE.value:
                return UserMemoryWrite.from_knx(raw)
            if apci == APCIUserService.USER_MANUFACTURER_INFO_READ.value:
                return UserManufacturerInfoRead.from_knx(raw)
            if apci == APCIUserService.USER_MANUFACTURER_INFO_RESPONSE.value:
                return UserManufacturerInfoResponse.from_knx(raw)
            if apci == APCIUserService.FUNCTION_PROPERTY_COMMAND.value:
                return FunctionPropertyCommand.from_knx(raw)
            if apci == APCIUserService.FUNCTION_PROPERTY_STATE_READ.value:
                return FunctionPropertyStateRead.from_knx(raw)
            if apci == APCIUserService.FUNCTION_PROPERTY_STATE_RESPONSE.value:
                return FunctionPropertyStateResponse.from_knx(raw)
        if service == APCIService.DEVICE_DESCRIPTOR_READ.value:
            return DeviceDescriptorRead.from_knx(raw)
        if service == APCIService.DEVICE_DESCRIPTOR_RESPONSE.value:
            return DeviceDescriptorResponse.from_knx(raw)
        if service == APCIService.RESTART.value:
            return Restart.from_knx(raw)
        if service == APCIService.ESCAPE.value:
            if apci == APCIExtendedService.AUTHORIZE_REQUEST.value:
                return AuthorizeRequest.from_knx(raw)
            if apci == APCIExtendedService.AUTHORIZE_RESPONSE.value:
                return AuthorizeResponse.from_knx(raw)
            if apci == APCIExtendedService.PROPERTY_VALUE_READ.value:
                return PropertyValueRead.from_knx(raw)
            if apci == APCIExtendedService.PROPERTY_VALUE_WRITE.value:
                return PropertyValueWrite.from_knx(raw)
            if apci == APCIExtendedService.PROPERTY_VALUE_RESPONSE.value:
                return PropertyValueResponse.from_knx(raw)
            if apci == APCIExtendedService.PROPERTY_DESCRIPTION_READ.value:
                return PropertyDescriptionRead.from_knx(raw)
            if apci == APCIExtendedService.PROPERTY_DESCRIPTION_RESPONSE.value:
                return PropertyDescriptionResponse.from_knx(raw)
            if apci == APCIExtendedService.INDIVIDUAL_ADDRESS_SERIAL_READ.value:
                return IndividualAddressSerialRead.from_knx(raw)
            if apci == APCIExtendedService.INDIVIDUAL_ADDRESS_SERIAL_RESPONSE.value:
                return IndividualAddressSerialResponse.from_knx(raw)
            if apci == APCIExtendedService.INDIVIDUAL_ADDRESS_SERIAL_WRITE.value:
                return IndividualAddressSerialWrite.from_knx(raw)
            if apci == APCIExtendedService.APCI_SEC.value:
                return SecureAPDU.from_knx(raw)

        raise ConversionError(f"Class not implemented for APCI {apci:#012b}.")


@dataclass(slots=True)
class GroupValueRead(APCI):
    """
    GroupValueRead service.

    Does not have any payload.
    """

    CODE: ClassVar = APCIService.GROUP_READ

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 1

    @classmethod
    def from_knx(cls, raw: bytes) -> GroupValueRead:
        """Parse/deserialize from KNX/IP raw data."""
        # Nothing to parse, but must be implemented explicitly.
        return cls()

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        return encode_cmd_and_payload(self.CODE)

    def __str__(self) -> str:
        """Return object as readable string."""
        return "<GroupValueRead />"


@dataclass(slots=True)
class GroupValueWrite(APCI):
    """
    GroupValueRead service.

    Takes a value (DPT) as payload.
    """

    CODE: ClassVar = APCIService.GROUP_WRITE

    value: DPTBinary | DPTArray

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        if isinstance(self.value, DPTBinary):
            return 1
        return 1 + len(self.value.value)

    @classmethod
    def from_knx(cls, raw: bytes) -> GroupValueWrite:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) == 2:
            return cls(value=DPTBinary(raw[1] & DPTBinary.APCI_BITMASK))

        return cls(value=DPTArray(raw[2:]))

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if isinstance(self.value, DPTBinary):
            return encode_cmd_and_payload(self.CODE, encoded_payload=self.value.value)

        return encode_cmd_and_payload(
            self.CODE, appended_payload=bytes(self.value.value)
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<GroupValueWrite value="{self.value}" />'


@dataclass(slots=True)
class GroupValueResponse(APCI):
    """
    GroupValueResponse service.

    Takes a value (DPT) as payload.
    """

    CODE: ClassVar = APCIService.GROUP_RESPONSE

    value: DPTBinary | DPTArray

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        if isinstance(self.value, DPTBinary):
            return 1
        return 1 + len(self.value.value)

    @classmethod
    def from_knx(cls, raw: bytes) -> GroupValueResponse:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) == 2:
            return cls(value=DPTBinary(raw[1] & DPTBinary.APCI_BITMASK))
        return cls(value=DPTArray(raw[2:]))

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if isinstance(self.value, DPTBinary):
            return encode_cmd_and_payload(self.CODE, encoded_payload=self.value.value)
        return encode_cmd_and_payload(
            self.CODE, appended_payload=bytes(self.value.value)
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<GroupValueResponse value="{self.value}" />'


@dataclass(slots=True)
class IndividualAddressWrite(APCI):
    """
    IndividualAddressWrite service.

    Payload contains the serial number and (new) address of the device.
    """

    CODE: ClassVar = APCIService.INDIVIDUAL_ADDRESS_WRITE

    address: IndividualAddress

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 3

    @classmethod
    def from_knx(cls, raw: bytes) -> IndividualAddressWrite:
        """Parse/deserialize from KNX/IP raw data."""
        (raw_address,) = struct.unpack("!H", raw[2:])
        return cls(address=IndividualAddress(raw_address))

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        return encode_cmd_and_payload(self.CODE, appended_payload=self.address.to_knx())

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<IndividualAddressWrite address="{self.address}" />'


@dataclass(slots=True)
class IndividualAddressRead(APCI):
    """IndividualAddressRead service."""

    CODE: ClassVar = APCIService.INDIVIDUAL_ADDRESS_READ

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 1

    @classmethod
    def from_knx(cls, raw: bytes) -> IndividualAddressRead:
        """Parse/deserialize from KNX/IP raw data."""
        # Nothing to parse, but must be implemented explicitly.
        return cls()

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        return encode_cmd_and_payload(self.CODE)

    def __str__(self) -> str:
        """Return object as readable string."""
        return "<IndividualAddressRead />"


@dataclass(slots=True)
class IndividualAddressResponse(APCI):
    """
    IndividualAddressResponse service.

    There is no payload, since the Telegram's source address is used as a
    response address.
    """

    CODE: ClassVar = APCIService.INDIVIDUAL_ADDRESS_RESPONSE

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 1

    @classmethod
    def from_knx(cls, raw: bytes) -> IndividualAddressResponse:
        """Parse/deserialize from KNX/IP raw data."""
        # Nothing to parse, but must be implemented explicitly.
        return cls()

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        return encode_cmd_and_payload(self.CODE)

    def __str__(self) -> str:
        """Return object as readable string."""
        return "<IndividualAddressResponse />"


@dataclass(slots=True)
class ADCRead(APCI):
    """
    ADCRead service.

    Payload contains the channel and number of samples to take.
    """

    CODE: ClassVar = APCIService.ADC_READ

    channel: int
    count: int = 1

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 2

    @classmethod
    def from_knx(cls, raw: bytes) -> ADCRead:
        """Parse/deserialize from KNX/IP raw data."""
        channel, count = struct.unpack("!BB", raw[1:])

        return cls(
            channel=channel & DPTBinary.APCI_BITMASK,
            count=count,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        payload = struct.pack("!BB", self.channel, self.count)

        return encode_cmd_and_payload(
            self.CODE, encoded_payload=payload[0], appended_payload=payload[1:]
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<ADCRead channel="{self.channel}" count="{self.count}" />'


@dataclass(slots=True)
class ADCResponse(APCI):
    """
    ADCResponse service.

    Payload contains the channel, number of samples and value.
    """

    CODE: ClassVar = APCIService.ADC_RESPONSE

    channel: int
    count: int = 1
    value: int = 0

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 4

    @classmethod
    def from_knx(cls, raw: bytes) -> ADCResponse:
        """Parse/deserialize from KNX/IP raw data."""
        channel, count, value = struct.unpack("!BBH", raw[1:])

        return cls(
            channel=channel & DPTBinary.APCI_BITMASK,
            count=count,
            value=value,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        payload = struct.pack("!BBH", self.channel, self.count, self.value)

        return encode_cmd_and_payload(
            self.CODE, encoded_payload=payload[0], appended_payload=payload[1:]
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<ADCResponse channel="{self.channel}" count="{self.count}" value="{self.value}" />'


@dataclass(slots=True)
class MemoryExtendedWrite(APCI):
    """
    MemoryExtendedWrite service.

    Payload indicates address (16 MiB), count (1-255 bytes) and data.
    """

    CODE: ClassVar = APCIService.MEMORY_EXTENDED_WRITE

    address: int
    data: bytes
    count: int = None  # type: ignore[assignment]

    def __post_init__(self) -> None:
        """Post-initialization steps."""
        if self.count is None:
            self.count = len(self.data)  # type: ignore[unreachable]

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 5 + len(self.data)

    @classmethod
    def from_knx(cls, raw: bytes) -> MemoryExtendedWrite:
        """Parse/deserialize from KNX/IP raw data."""
        count = raw[2]
        address = int.from_bytes(raw[3:6], "big")
        data = raw[6:]

        return cls(
            count=count,
            address=address,
            data=data,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.address <= 0xFFFFFF:
            raise ConversionError("Address out of range.")
        if not 0 <= self.count <= 250:
            raise ConversionError("Count out of range.")

        size = len(self.data)
        payload = struct.pack(f"!BI{size}s", self.count, self.address, self.data)
        # suppress first byte of address
        payload = payload[:1] + payload[2:]

        return encode_cmd_and_payload(self.CODE, 0, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<MemoryExtendedWrite address="{hex(self.address)}" count="{self.count}" data="{self.data.hex()}" />'


@dataclass(slots=True)
class MemoryExtendedWriteResponse(APCI):
    """
    MemoryExtendedWriteResponse service.

    Payload indicates return code, address (16 MiB) and confirmation data.
    """

    CODE: ClassVar = APCIService.MEMORY_EXTENDED_WRITE_RESPONSE

    return_code: int
    address: int
    confirmation_data: bytes = b""

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 5 + len(self.confirmation_data)

    @classmethod
    def from_knx(cls, raw: bytes) -> MemoryExtendedWriteResponse:
        """Parse/deserialize from KNX/IP raw data."""
        return_code = raw[2]
        address = int.from_bytes(raw[3:6], "big")
        confirmation_data = raw[6:]

        return cls(
            return_code=return_code,
            address=address,
            confirmation_data=confirmation_data,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.address <= 0xFFFFFF:
            raise ConversionError("Address out of range.")
        if not 0 <= self.return_code <= 255:
            raise ConversionError("Return code out of range.")

        size = len(self.confirmation_data)
        payload = struct.pack(
            f"!BI{size}s", self.return_code, self.address, self.confirmation_data
        )
        # suppress first byte of address
        payload = payload[:1] + payload[2:]

        return encode_cmd_and_payload(self.CODE, 0, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<MemoryExtendedWriteResponse return_code="{self.return_code}" address="{hex(self.address)}" confirmation_data="{self.confirmation_data.hex()}" />'


@dataclass(slots=True)
class MemoryExtendedRead(APCI):
    """
    MemoryExtendedRead service.

    Payload indicates count and address (16 MiB).
    """

    CODE: ClassVar = APCIService.MEMORY_EXTENDED_READ

    count: int
    address: int

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 5

    @classmethod
    def from_knx(cls, raw: bytes) -> MemoryExtendedRead:
        """Parse/deserialize from KNX/IP raw data."""
        count = raw[2]
        address = int.from_bytes(raw[3:6], "big")

        return cls(
            count=count,
            address=address,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.address <= 0xFFFFFF:
            raise ConversionError("Address out of range.")
        if not 0 <= self.count <= 250:
            raise ConversionError("Count out of range.")

        payload = struct.pack("!BI", self.count, self.address)
        # suppress first byte of address
        payload = payload[:1] + payload[2:]

        return encode_cmd_and_payload(self.CODE, 0, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            f'<MemoryExtendedRead count="{self.count}" address="{hex(self.address)}" />'
        )


@dataclass(slots=True)
class MemoryExtendedReadResponse(APCI):
    """
    MemoryExtendedReadResponse service.

    Payload indicates return code, address (16 MiB) and data.
    """

    CODE: ClassVar = APCIService.MEMORY_EXTENDED_READ_RESPONSE

    return_code: int
    address: int
    data: bytes = b""

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 5 + len(self.data)

    @classmethod
    def from_knx(cls, raw: bytes) -> MemoryExtendedReadResponse:
        """Parse/deserialize from KNX/IP raw data."""
        return_code = raw[2]
        address = int.from_bytes(raw[3:6], "big")
        data = raw[6:]

        return cls(
            return_code=return_code,
            address=address,
            data=data,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.address <= 0xFFFFFF:
            raise ConversionError("Address out of range.")

        if not 0 <= self.return_code <= 255:
            raise ConversionError("Return code out of range.")

        size = len(self.data)
        payload = struct.pack(f"!BI{size}s", self.return_code, self.address, self.data)
        # suppress first byte of address
        payload = payload[:1] + payload[2:]

        return encode_cmd_and_payload(self.CODE, 0, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<MemoryExtendedReadResponse return_code="{self.return_code}" address="{hex(self.address)}" data="{self.data.hex()}" />'


@dataclass(slots=True)
class MemoryRead(APCI):
    """
    MemoryRead service.

    Payload indicates address (64 KiB) and count (1-63 bytes).
    """

    CODE: ClassVar = APCIService.MEMORY_READ

    address: int
    count: int = 1

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 3

    @classmethod
    def from_knx(cls, raw: bytes) -> MemoryRead:
        """Parse/deserialize from KNX/IP raw data."""
        count, address = struct.unpack("!BH", raw[1:])

        return cls(
            address=address,
            count=count & DPTBinary.APCI_BITMASK,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.address <= 0xFFFF:
            raise ConversionError("Address out of range.")
        if not 0 <= self.count <= 0x3F:
            raise ConversionError("Count out of range.")

        payload = struct.pack("!BH", self.count, self.address)

        return encode_cmd_and_payload(
            self.CODE, encoded_payload=payload[0], appended_payload=payload[1:]
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<MemoryRead address="{hex(self.address)}" count="{self.count}" />'


@dataclass(slots=True)
class MemoryWrite(APCI):
    """
    MemoryWrite service.

    Payload indicates address (64 KiB), count (1-63 bytes) and data.
    """

    CODE: ClassVar = APCIService.MEMORY_WRITE

    address: int
    data: bytes
    count: int = None  # type: ignore[assignment]

    def __post_init__(self) -> None:
        """Post-initialization steps."""
        if self.count is None:
            self.count = len(self.data)  # type: ignore[unreachable]

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 3 + len(self.data)

    @classmethod
    def from_knx(cls, raw: bytes) -> MemoryWrite:
        """Parse/deserialize from KNX/IP raw data."""
        size = len(raw) - 4

        count, address, data = struct.unpack(f"!BH{size}s", raw[1:])

        return cls(
            count=count & DPTBinary.APCI_BITMASK,
            address=address,
            data=data,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.address <= 0xFFFF:
            raise ConversionError("Address out of range.")
        if not 0 <= self.count <= 0x3F:
            raise ConversionError("Count out of range.")

        size = len(self.data)
        payload = struct.pack(f"!BH{size}s", self.count, self.address, self.data)

        return encode_cmd_and_payload(
            self.CODE, encoded_payload=payload[0], appended_payload=payload[1:]
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<MemoryWrite address="{hex(self.address)}" count="{self.count}" data="{self.data.hex()}" />'


@dataclass(slots=True)
class MemoryResponse(APCI):
    """
    MemoryResponse service.

    Payload indicates address (64 KiB), count (1-63 bytes) and data.
    """

    CODE: ClassVar = APCIService.MEMORY_RESPONSE

    address: int
    data: bytes
    count: int = None  # type: ignore[assignment]

    def __post_init__(self) -> None:
        """Post-initialization steps."""
        if self.count is None:
            self.count = len(self.data)  # type: ignore[unreachable]

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 3 + len(self.data)

    @classmethod
    def from_knx(cls, raw: bytes) -> MemoryResponse:
        """Parse/deserialize from KNX/IP raw data."""
        size = len(raw) - 4

        count, address, data = struct.unpack(f"!BH{size}s", raw[1:])

        return cls(
            count=count & DPTBinary.APCI_BITMASK,
            address=address,
            data=data,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.address <= 0xFFFF:
            raise ConversionError("Address out of range.")
        if not 0 <= self.count <= 0x3F:
            raise ConversionError("Count out of range.")

        size = len(self.data)
        payload = struct.pack(f"!BH{size}s", self.count, self.address, self.data)

        return encode_cmd_and_payload(
            self.CODE, encoded_payload=payload[0], appended_payload=payload[1:]
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<MemoryResponse address="{hex(self.address)}" count="{self.count}" data="{self.data.hex()}" />'


@dataclass(slots=True)
class DeviceDescriptorRead(APCI):
    """
    DeviceDescriptorRead service.

    Payload contains the descriptor.
    """

    CODE: ClassVar = APCIService.DEVICE_DESCRIPTOR_READ

    descriptor: int = 0

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 1

    @classmethod
    def from_knx(cls, raw: bytes) -> DeviceDescriptorRead:
        """Parse/deserialize from KNX/IP raw data."""
        return cls(descriptor=raw[1] & 0x3F)

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.descriptor <= 0x3F:
            raise ConversionError("Descriptor out of range.")

        return encode_cmd_and_payload(self.CODE, encoded_payload=self.descriptor)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<DeviceDescriptorRead descriptor="{self.descriptor}" />'


@dataclass(slots=True)
class DeviceDescriptorResponse(APCI):
    """
    DeviceDescriptorResponse service.

    Payload contains the descriptor and value.
    """

    CODE: ClassVar = APCIService.DEVICE_DESCRIPTOR_RESPONSE

    descriptor: int = 0
    value: int = 0

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 3

    @classmethod
    def from_knx(cls, raw: bytes) -> DeviceDescriptorResponse:
        """Parse/deserialize from KNX/IP raw data."""
        descriptor, value = struct.unpack("!BH", raw[1:])

        return cls(descriptor=descriptor & 0x3F, value=value)

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.descriptor <= 0x3F:
            raise ConversionError("Descriptor out of range.")

        payload = struct.pack("!H", self.value)

        return encode_cmd_and_payload(
            self.CODE, encoded_payload=self.descriptor, appended_payload=payload
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<DeviceDescriptorResponse descriptor="{self.descriptor}" value="{self.value}" />'


@dataclass(slots=True)
class Restart(APCI):
    """
    Restart service.

    Does not take any payload.
    """

    # Requests a Basic Restart of the communication partner.
    # Master reset is not implemented yet.

    CODE: ClassVar = APCIService.RESTART

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 1

    @classmethod
    def from_knx(cls, raw: bytes) -> Restart:
        """Parse/deserialize from KNX/IP raw data."""
        # Nothing to parse, but must be implemented explicitly.
        return cls()

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        return encode_cmd_and_payload(self.CODE)

    def __str__(self) -> str:
        """Return object as readable string."""
        return "<Restart />"


@dataclass(slots=True)
class UserMemoryRead(APCI):
    """
    UserMemoryRead service.

    Payload indicates address (1 MiB) and count (1-15 bytes).
    """

    CODE: ClassVar = APCIUserService.USER_MEMORY_READ

    address: int = 0
    count: int = 1

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 4

    @classmethod
    def from_knx(cls, raw: bytes) -> UserMemoryRead:
        """Parse/deserialize from KNX/IP raw data."""
        byte0, address = struct.unpack("!BH", raw[2:])

        return cls(
            count=byte0 & 0x0F,
            address=(((byte0 & 0xF0) >> 4) << 16) + address,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.address <= 0xFFFFF:
            raise ConversionError("Address out of range.")
        if not 0 <= self.count <= 0xF:
            raise ConversionError("Count out of range.")

        byte0 = (((self.address & 0x0F0000) >> 16) << 4) | (self.count & 0x0F)
        address = self.address & 0xFFFF

        payload = struct.pack("!BH", byte0, address)

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<UserMemoryRead address="{hex(self.address)}" count="{self.count}" />'


@dataclass(slots=True)
class UserMemoryWrite(APCI):
    """
    UserMemoryWrite service.

    Payload indicates address (1 MiB), count and data.
    """

    CODE: ClassVar = APCIUserService.USER_MEMORY_WRITE

    address: int
    data: bytes
    count: int = None  # type: ignore[assignment]

    def __post_init__(self) -> None:
        """Post-initialization steps."""
        if self.count is None:
            self.count = len(self.data)  # type: ignore[unreachable]

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 4 + len(self.data)

    @classmethod
    def from_knx(cls, raw: bytes) -> UserMemoryWrite:
        """Parse/deserialize from KNX/IP raw data."""
        size = len(raw) - 5

        byte0, address, data = struct.unpack(f"!BH{size}s", raw[2:])

        return cls(
            address=(((byte0 & 0xF0) >> 4) << 16) + address,
            data=data,
            count=byte0 & 0x0F,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.address <= 0xFFFFF:
            raise ConversionError("Address out of range.")
        if not 0 <= self.count <= 0xF:
            raise ConversionError("Count out of range.")

        byte0 = (((self.address & 0x0F0000) >> 16) << 4) | (self.count & 0x0F)
        address = self.address & 0xFFFF

        size = len(self.data)
        payload = struct.pack(f"!BH{size}s", byte0, address, self.data)

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<UserMemoryWrite address="{hex(self.address)}" count="{self.count}" data="{self.data.hex()}" />'


@dataclass(slots=True)
class UserMemoryResponse(APCI):
    """
    UserMemoryResponse service.

    Payload indicates address (1 MiB), count and data.
    """

    CODE: ClassVar = APCIUserService.USER_MEMORY_RESPONSE

    address: int
    data: bytes
    count: int = None  # type: ignore[assignment]

    def __post_init__(self) -> None:
        """Post-initialization steps."""
        if self.count is None:
            self.count = len(self.data)  # type: ignore[unreachable]

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 4 + len(self.data)

    @classmethod
    def from_knx(cls, raw: bytes) -> UserMemoryResponse:
        """Parse/deserialize from KNX/IP raw data."""
        size = len(raw) - 5

        byte0, address, data = struct.unpack(f"!BH{size}s", raw[2:])

        return cls(
            address=(((byte0 & 0xF0) >> 4) << 16) + address,
            data=data,
            count=byte0 & 0x0F,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.address <= 0xFFFFF:
            raise ConversionError("Address out of range.")
        if not 0 <= self.count <= 0xF:
            raise ConversionError("Count out of range.")

        byte0 = (((self.address & 0x0F0000) >> 16) << 4) | (self.count & 0x0F)
        address = self.address & 0xFFFF

        size = len(self.data)
        payload = struct.pack(f"!BH{size}s", byte0, address, self.data)

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<UserMemoryResponse address="{hex(self.address)}" count="{self.count}" data="{self.data.hex()}" />'


@dataclass(slots=True)
class UserManufacturerInfoRead(APCI):
    """UserManufacturerInfoRead service."""

    CODE: ClassVar = APCIUserService.USER_MANUFACTURER_INFO_READ

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 1

    @classmethod
    def from_knx(cls, raw: bytes) -> UserManufacturerInfoRead:
        """Parse/deserialize from KNX/IP raw data."""
        # Nothing to parse, but must be implemented explicitly.
        return cls()

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""

        return encode_cmd_and_payload(self.CODE)

    def __str__(self) -> str:
        """Return object as readable string."""
        return "<UserManufacturerInfoRead />"


@dataclass(slots=True)
class UserManufacturerInfoResponse(APCI):
    """UserManufacturerInfoResponse service."""

    CODE: ClassVar = APCIUserService.USER_MANUFACTURER_INFO_RESPONSE

    manufacturer_id: int = 0
    data: bytes = b"\x00\x00"

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 4

    @classmethod
    def from_knx(cls, raw: bytes) -> UserManufacturerInfoResponse:
        """Parse/deserialize from KNX/IP raw data."""
        manufacturer_id, data = struct.unpack("!B2s", raw[2:])
        return cls(manufacturer_id=manufacturer_id, data=data)

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        payload = struct.pack("!B2s", self.manufacturer_id, self.data)

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<UserManufacturerInfoResponse manufacturer_id="{self.manufacturer_id}" data="{self.data.hex()}" />'


@dataclass(slots=True)
class FunctionPropertyCommand(APCI):
    """FunctionPropertyCommand service."""

    CODE: ClassVar = APCIUserService.FUNCTION_PROPERTY_COMMAND

    object_index: int = 0
    property_id: int = 0
    data: bytes = b""

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 3 + len(self.data)

    @classmethod
    def from_knx(cls, raw: bytes) -> FunctionPropertyCommand:
        """Parse/deserialize from KNX/IP raw data."""
        size = len(raw) - 4
        object_index, property_id, data = struct.unpack(f"!BB{size}s", raw[2:])
        return cls(
            object_index=object_index,
            property_id=property_id,
            data=data,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        size = len(self.data)
        payload = struct.pack(
            f"!BB{size}s", self.object_index, self.property_id, self.data
        )

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<FunctionPropertyCommand object_index="{self.object_index}" property_id="{self.property_id}" data="{self.data.hex()}" />'


@dataclass(slots=True)
class FunctionPropertyStateRead(APCI):
    """FunctionPropertyStateRead service."""

    CODE: ClassVar = APCIUserService.FUNCTION_PROPERTY_STATE_READ

    object_index: int = 0
    property_id: int = 0
    data: bytes = b""

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 3 + len(self.data)

    @classmethod
    def from_knx(cls, raw: bytes) -> FunctionPropertyStateRead:
        """Parse/deserialize from KNX/IP raw data."""
        size = len(raw) - 4
        object_index, property_id, data = struct.unpack(f"!BB{size}s", raw[2:])
        return cls(
            object_index=object_index,
            property_id=property_id,
            data=data,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        size = len(self.data)
        payload = struct.pack(
            f"!BB{size}s", self.object_index, self.property_id, self.data
        )

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<FunctionPropertyStateRead object_index="{self.object_index}" property_id="{self.property_id}" data="{self.data.hex()}" />'


@dataclass(slots=True)
class FunctionPropertyStateResponse(APCI):
    """FunctionPropertyStateResponse service."""

    CODE: ClassVar = APCIUserService.FUNCTION_PROPERTY_STATE_RESPONSE

    object_index: int = 0
    property_id: int = 0
    return_code: int = 0
    data: bytes = b""

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 4 + len(self.data)

    @classmethod
    def from_knx(cls, raw: bytes) -> FunctionPropertyStateResponse:
        """Parse/deserialize from KNX/IP raw data."""
        size = len(raw) - 5
        (
            object_index,
            property_id,
            return_code,
            data,
        ) = struct.unpack(f"!BBB{size}s", raw[2:])
        return cls(
            object_index=object_index,
            property_id=property_id,
            return_code=return_code,
            data=data,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        size = len(self.data)
        payload = struct.pack(
            f"!BBB{size}s",
            self.object_index,
            self.property_id,
            self.return_code,
            self.data,
        )

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<FunctionPropertyStateResponse object_index="{self.object_index}" property_id="{self.property_id}" return_code="{self.return_code}" data="{self.data.hex()}" />'


@dataclass(slots=True)
class AuthorizeRequest(APCI):
    """AuthorizeRequest service."""

    CODE: ClassVar = APCIExtendedService.AUTHORIZE_REQUEST

    key: int = 0

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 6

    @classmethod
    def from_knx(cls, raw: bytes) -> AuthorizeRequest:
        """Parse/deserialize from KNX/IP raw data."""
        _, key = struct.unpack("!BI", raw[2:])
        return cls(key=key)

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        payload = struct.pack("!BI", 0, self.key)

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<AuthorizeRequest key="{self.key}" />'


@dataclass(slots=True)
class AuthorizeResponse(APCI):
    """AuthorizeResponse service."""

    CODE: ClassVar = APCIExtendedService.AUTHORIZE_RESPONSE

    level: int = 0

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 2

    @classmethod
    def from_knx(cls, raw: bytes) -> AuthorizeResponse:
        """Parse/deserialize from KNX/IP raw data."""
        (level,) = struct.unpack("!B", raw[2:])
        return cls(level=level)

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        payload = struct.pack("!B", self.level)

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<AuthorizeResponse level="{self.level}"/>'


@dataclass(slots=True)
class PropertyValueRead(APCI):
    """
    PropertyValueRead service.

    Payload indicates object, property, count and start.
    """

    CODE: ClassVar = APCIExtendedService.PROPERTY_VALUE_READ

    object_index: int = 0
    property_id: int = 0
    count: int = 1
    start_index: int = 1

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 5

    @classmethod
    def from_knx(cls, raw: bytes) -> PropertyValueRead:
        """Parse/deserialize from KNX/IP raw data."""
        (
            object_index,
            property_id,
            count,
            start_index,
        ) = struct.unpack("!BBBB", raw[2:])
        return cls(
            object_index=object_index,
            property_id=property_id,
            count=count >> 4,
            start_index=(count & 0xF) * 256 + start_index,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.count <= 0xF:
            raise ConversionError("Count out of range.")

        payload = struct.pack(
            "!BBBB",
            self.object_index,
            self.property_id,
            (self.count << 4) + (self.start_index >> 8),
            self.start_index & 0xFF,
        )

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            "<PropertyValueRead "
            f'object_index="{self.object_index}" '
            f'property_id="{self.property_id}" '
            f'count="{self.count}" '
            f'start_index="{self.start_index}" '
            "/>"
        )


@dataclass(slots=True)
class PropertyValueWrite(APCI):
    """
    PropertyValueWrite service.

    Payload indicates object, property, count, start and data itself.
    """

    CODE: ClassVar = APCIExtendedService.PROPERTY_VALUE_WRITE

    object_index: int = 0
    property_id: int = 0
    count: int = 1
    start_index: int = 0
    data: bytes = b""

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 5 + len(self.data)

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.count <= 0xF:
            raise ConversionError("Count out of range.")

        size = len(self.data)
        payload = struct.pack(
            f"!BBBB{size}s",
            self.object_index,
            self.property_id,
            (self.count << 4) + (self.start_index >> 8),
            self.start_index & 0xFF,
            self.data,
        )

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    @classmethod
    def from_knx(cls, raw: bytes) -> PropertyValueWrite:
        """Parse/deserialize from KNX/IP raw data."""
        size = len(raw) - 6
        (
            object_index,
            property_id,
            count,
            start_index,
            data,
        ) = struct.unpack(f"!BBBB{size}s", raw[2:])
        return cls(
            object_index=object_index,
            property_id=property_id,
            count=count >> 4,
            start_index=(count & 0xF) * 256 + start_index,
            data=data,
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            "<PropertyValueWrite "
            f'object_index="{self.object_index}" '
            f'property_id="{self.property_id}" '
            f'count="{self.count}" '
            f'start_index="{self.start_index}" '
            f'data="{self.data.hex()}" '
            "/>"
        )


@dataclass(slots=True)
class PropertyValueResponse(APCI):
    """
    PropertyValueResponse service.

    Payload indicates object, property, count, start and data itself. Size of
    the payload depends on the data.
    """

    CODE: ClassVar = APCIExtendedService.PROPERTY_VALUE_RESPONSE

    object_index: int = 0
    property_id: int = 0
    count: int = 1
    start_index: int = 0
    data: bytes = b""

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 5 + len(self.data)

    @classmethod
    def from_knx(cls, raw: bytes) -> PropertyValueResponse:
        """Parse/deserialize from KNX/IP raw data."""
        size = len(raw) - 6
        (
            object_index,
            property_id,
            count,
            start_index,
            data,
        ) = struct.unpack(f"!BBBB{size}s", raw[2:])
        return cls(
            object_index=object_index,
            property_id=property_id,
            count=count >> 4,
            start_index=(count & 0xF) * 256 + start_index,
            data=data,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.count <= 0xF:
            raise ConversionError("Count out of range.")

        size = len(self.data)
        payload = struct.pack(
            f"!BBBB{size}s",
            self.object_index,
            self.property_id,
            (self.count << 4) + (self.start_index >> 8),
            self.start_index & 0xFF,
            self.data,
        )

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            "<PropertyValueResponse "
            f'object_index="{self.object_index}" '
            f'property_id="{self.property_id}" '
            f'count="{self.count}" '
            f'start_index="{self.start_index}" '
            f'data="{self.data.hex()}" '
            "/>"
        )


@dataclass(slots=True)
class PropertyDescriptionRead(APCI):
    """PropertyDescriptionRead service."""

    CODE: ClassVar = APCIExtendedService.PROPERTY_DESCRIPTION_READ

    object_index: int = 0
    property_id: int = 0
    property_index: int = 0

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 4

    @classmethod
    def from_knx(cls, raw: bytes) -> PropertyDescriptionRead:
        """Parse/deserialize from KNX/IP raw data."""
        object_index, property_id, property_index = struct.unpack("!BBB", raw[2:])
        return cls(
            object_index=object_index,
            property_id=property_id,
            property_index=property_index,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        payload = struct.pack(
            "!BBB", self.object_index, self.property_id, self.property_index
        )

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<PropertyDescriptionRead object_index="{self.object_index}" property_id="{self.property_id}" property_index="{self.property_index}" />'


@dataclass(slots=True)
class PropertyDescriptionResponse(APCI):
    """PropertyDescriptionResponse service."""

    CODE: ClassVar = APCIExtendedService.PROPERTY_DESCRIPTION_RESPONSE

    object_index: int = 0
    property_id: int = 0
    property_index: int = 0
    type_: int = 0
    max_count: int = 1
    access: int = 0

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 8

    @classmethod
    def from_knx(cls, raw: bytes) -> PropertyDescriptionResponse:
        """Parse/deserialize from KNX/IP raw data."""
        (
            object_index,
            property_id,
            property_index,
            type_,
            max_count,
            access,
        ) = struct.unpack("!BBBBHB", raw[2:])

        return cls(
            object_index=object_index,
            property_id=property_id,
            property_index=property_index,
            type_=type_,
            max_count=max_count & 0x0FFF,
            access=access,
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if not 0 <= self.max_count <= 0x0FFF:
            raise ConversionError("Max count out of range.")

        payload = struct.pack(
            "!BBBBHB",
            self.object_index,
            self.property_id,
            self.property_index,
            self.type_,
            self.max_count & 0x0FFF,
            self.access,
        )

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<PropertyDescriptionResponse object_index="{self.object_index}" property_id="{self.property_id}" property_index="{self.property_index}" type="{self.type_}" max_count="{self.max_count}" access="{self.access}" />'


@dataclass(slots=True)
class IndividualAddressSerialRead(APCI):
    """IndividualAddressSerialRead service."""

    CODE: ClassVar = APCIExtendedService.INDIVIDUAL_ADDRESS_SERIAL_READ

    serial: bytes

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 7

    @classmethod
    def from_knx(cls, raw: bytes) -> IndividualAddressSerialRead:
        """Parse/deserialize from KNX/IP raw data."""
        (serial,) = struct.unpack("!6s", raw[2:])
        return cls(serial=serial)

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if len(self.serial) != 6:
            raise ConversionError("Serial must be 6 bytes.")

        payload = struct.pack("!6s", self.serial)

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<IndividualAddressSerialRead serial="{self.serial.hex()}" />'


@dataclass(slots=True)
class IndividualAddressSerialResponse(APCI):
    """IndividualAddressSerialResponse service."""

    CODE: ClassVar = APCIExtendedService.INDIVIDUAL_ADDRESS_SERIAL_RESPONSE

    serial: bytes
    address: IndividualAddress

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 11

    @classmethod
    def from_knx(cls, raw: bytes) -> IndividualAddressSerialResponse:
        """Parse/deserialize from KNX/IP raw data."""
        serial, raw_address, _ = struct.unpack("!6sHH", raw[2:])
        return cls(
            serial=serial,
            address=IndividualAddress(raw_address),
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if len(self.serial) != 6:
            raise ConversionError("Serial must be 6 bytes.")

        payload = struct.pack("!6s2sH", self.serial, self.address.to_knx(), 0)

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<IndividualAddressSerialResponse serial="{self.serial.hex()}" address="{self.address}" />'


@dataclass(slots=True)
class IndividualAddressSerialWrite(APCI):
    """IndividualAddressSerialWrite service."""

    CODE: ClassVar = APCIExtendedService.INDIVIDUAL_ADDRESS_SERIAL_WRITE

    serial: bytes
    address: IndividualAddress

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 13

    @classmethod
    def from_knx(cls, raw: bytes) -> IndividualAddressSerialWrite:
        """Parse/deserialize from KNX/IP raw data."""
        serial, raw_address, _ = struct.unpack("!6sHI", raw[2:])
        return cls(
            serial=serial,
            address=IndividualAddress(raw_address),
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        if len(self.serial) != 6:
            raise ConversionError("Serial must be 6 bytes.")

        payload = struct.pack("!6s2sI", self.serial, self.address.to_knx(), 0)

        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<IndividualAddressSerialWrite serial="{self.serial.hex()}" address="{self.address}" />'


@dataclass(slots=True)
class SecureAPDU(APCI):
    """SecureAPDU service."""

    CODE: ClassVar = APCIExtendedService.APCI_SEC

    scf: SecurityControlField
    secured_data: SecureData

    def calculated_length(self) -> int:
        """Get length of APCI payload."""
        return 2 + len(self.secured_data)

    @classmethod
    def from_knx(cls, raw: bytes) -> SecureAPDU:
        """Parse/deserialize from KNX/IP raw data."""
        return cls(
            scf=SecurityControlField.from_knx(raw[2]),
            secured_data=SecureData.from_knx(raw[3:]),
        )

    def to_knx(self) -> bytearray:
        """Serialize to KNX/IP raw data."""
        payload = self.scf.to_knx() + self.secured_data.to_knx()
        return encode_cmd_and_payload(self.CODE, appended_payload=payload)

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<SecureAPDU scf="{self.scf}" secured_data={self.secured_data!r} />'



"""Unit test for KNX/IP SearchRequest objects."""

from xknx.knxip import HPAI, KNXIPFrame, SearchRequest


class TestKNXIPSearchRequest:
    """Test class for KNX/IP SearchRequest objects."""

    def test_search_request(self) -> None:
        """Test parsing and streaming SearchRequest KNX/IP packet."""
        raw = bytes.fromhex("06 10 02 01 00 0E 08 01 E0 00 17 0C 0E 57")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, SearchRequest)
        assert knxipframe.body.discovery_endpoint == HPAI(
            ip_addr="224.0.23.12", port=3671
        )

        knxipframe2 = KNXIPFrame.init_from_body(
            SearchRequest(discovery_endpoint=HPAI(ip_addr="224.0.23.12", port=3671))
        )

        assert knxipframe2.to_knx() == raw



"""Implementation of KNX RGB color data point type."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any

from .dpt import DPTComplex, DPTComplexData
from .payload import DPTArray, DPTBinary


@dataclass(slots=True)
class RGBColor(DPTComplexData):
    """
    Representation of RGB color.

    `red`: int 0..255
    `green`: int 0..255
    `blue`: int 0..255
    """

    red: int
    green: int
    blue: int

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> RGBColor:
        """Init from a dictionary."""
        try:
            red = int(data["red"])
            green = int(data["green"])
            blue = int(data["blue"])
        except KeyError as err:
            raise ValueError(f"Missing required key: {err}") from err
        except (ValueError, TypeError) as err:
            raise ValueError(f"Invalid value for RGB color: {err}") from err

        return cls(red=red, green=green, blue=blue)

    def as_dict(self) -> dict[str, int]:
        """Create a JSON serializable dictionary."""
        return {
            "red": self.red,
            "green": self.green,
            "blue": self.blue,
        }


class DPTColorRGB(DPTComplex[RGBColor]):
    """Abstraction for KNX 3 octet color RGB (DPT 232.600)."""

    data_type = RGBColor
    payload_type = DPTArray
    payload_length = 3
    dpt_main_number = 232
    dpt_sub_number = 600
    value_type = "color_rgb"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> RGBColor:
        """Parse/deserialize from KNX/IP raw data."""
        raw = cls.validate_payload(payload)
        return RGBColor(
            red=raw[0],
            green=raw[1],
            blue=raw[2],
        )

    @classmethod
    def _to_knx(cls, value: RGBColor) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        colors: tuple[int, int, int] = (value.red, value.green, value.blue)
        for color in colors:
            if not 0 <= color <= 255:
                raise ValueError(f"Color value out of range 0..255: {value}")
        return DPTArray(colors)



"""Unit test for KNX DPT 1 byte relative value  objects."""

import pytest

from xknx.dpt import DPTArray, DPTPercentV8, DPTSignedRelativeValue, DPTValue1Count
from xknx.exceptions import ConversionError


class TestDPTRelativeValue:
    """Test class for KNX DPT Relative Value."""

    @pytest.mark.parametrize(
        ("raw", "value"),
        [
            ((0x00,), 0),
            ((0x01,), 1),
            ((0x02,), 2),
            ((0x64,), 100),
            ((0x7F,), 127),
            ((0x80,), -128),
            ((0x9C,), -100),
            ((0xFE,), -2),
            ((0xFF,), -1),
        ],
    )
    def test_transcoder(self, raw: tuple[int], value: int) -> None:
        """Test value from and to KNX."""
        assert DPTSignedRelativeValue.from_knx(DPTArray(raw)) == value
        assert DPTSignedRelativeValue.to_knx(value) == DPTArray(raw)

    def test_assert_min_exceeded(self) -> None:
        """Test initialization with wrong value (Underflow)."""
        with pytest.raises(ConversionError):
            DPTSignedRelativeValue.to_knx(-129)

    def test_assert_max_exceeded(self) -> None:
        """Test initialization with wrong value (Overflow)."""
        with pytest.raises(ConversionError):
            DPTSignedRelativeValue.to_knx(128)

    def test_unit(self) -> None:
        """Test unit of 1 byte relative value objects."""
        assert DPTSignedRelativeValue.unit is None
        assert DPTPercentV8.unit == "%"
        assert DPTValue1Count.unit == "counter pulses"



"""Module for managing a KNX scene."""

from __future__ import annotations

from collections.abc import Iterator
import logging
from typing import TYPE_CHECKING

from xknx.remote_value import GroupAddressesType, RemoteValueSceneNumber

from .device import Device, DeviceCallbackType

if TYPE_CHECKING:
    from xknx.telegram import Telegram
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.log")


class Scene(Device):
    """Class for managing a scene."""

    def __init__(
        self,
        xknx: XKNX,
        name: str,
        group_address: GroupAddressesType = None,
        scene_number: int = 1,
        device_updated_cb: DeviceCallbackType[Scene] | None = None,
    ) -> None:
        """Initialize Sceneclass."""
        super().__init__(xknx, name, device_updated_cb)

        # TODO: state_updater: disable for scene number per default?
        self.scene_value = RemoteValueSceneNumber(
            xknx,
            group_address=group_address,
            device_name=self.name,
            feature_name="Scene number",
            after_update_cb=self._scene_activation_from_rv,
        )
        self.scene_number = int(scene_number)

    def _iter_remote_values(self) -> Iterator[RemoteValueSceneNumber]:
        """Iterate the devices RemoteValue classes."""
        yield self.scene_value

    async def run(self) -> None:
        """Activate scene."""
        self.scene_value.set(self.scene_number)

    def process_group_write(self, telegram: Telegram) -> None:
        """Process incoming and outgoing GROUP WRITE telegram."""
        self.scene_value.process(telegram, always_callback=True)

    def _scene_activation_from_rv(self, called_scene_number: int) -> None:
        """Check the scene number from RemoteValue (Callback)."""
        if called_scene_number == self.scene_number:
            self.after_update()

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            f'<Scene name="{self.name}" '
            f"scene_value={self.scene_value.group_addr_str()} "
            f'scene_number="{self.scene_number}" />'
        )



"""
Module for Serialization and Deserialization of a KNX Connect Request information.

Connect requests are used to start a new tunnel connection on a KNX/IP device.
"""

from __future__ import annotations

from xknx.exceptions import CouldNotParseKNXIP
from xknx.telegram import IndividualAddress

from .body import KNXIPBody
from .hpai import HPAI
from .knxip_enum import ConnectRequestType, KNXIPServiceType, TunnellingLayer


class ConnectRequest(KNXIPBody):
    """Representation of a KNX Connect Request."""

    SERVICE_TYPE = KNXIPServiceType.CONNECT_REQUEST

    def __init__(
        self,
        control_endpoint: HPAI | None = None,
        data_endpoint: HPAI | None = None,
        cri: ConnectRequestInformation | None = None,
    ) -> None:
        """Initialize ConnectRequest object."""
        self.control_endpoint = control_endpoint or HPAI()
        self.data_endpoint = data_endpoint or HPAI()
        self.cri = cri or ConnectRequestInformation()

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return HPAI.LENGTH + HPAI.LENGTH + self.cri.calculated_length()

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        pos = self.control_endpoint.from_knx(raw)
        pos += self.data_endpoint.from_knx(raw[pos:])
        pos += self.cri.from_knx(raw[pos:])
        return pos

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return (
            self.control_endpoint.to_knx()
            + self.data_endpoint.to_knx()
            + self.cri.to_knx()
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<ConnectRequest "
            f'control_endpoint="{self.control_endpoint}" '
            f'data_endpoint="{self.data_endpoint}" '
            f'cri="{self.cri}" />'
        )


class ConnectRequestInformation:
    """
    Representation of a KNX Connect Request Information (CRI).

    A Basic CRI requests a tunnel without requesting any specific IA.
    Using `individual_address` yields an Extended CRI which is only
    supported by Tunnelling v2 devices.
    """

    CRI_LENGTH = 2
    CRI_TUNNEL_LENGTH = 4
    CRI_TUNNEL_EXT_LENGTH = 6

    def __init__(
        self,
        connection_type: ConnectRequestType = ConnectRequestType.TUNNEL_CONNECTION,
        knx_layer: TunnellingLayer = TunnellingLayer.DATA_LINK_LAYER,
        individual_address: IndividualAddress | None = None,
    ) -> None:
        """Initialize ConnectRequest object."""
        self.connection_type = connection_type
        self.knx_layer = knx_layer
        self.individual_address = individual_address

    def _is_tunnel_cri(self) -> bool:
        return self.connection_type == ConnectRequestType.TUNNEL_CONNECTION

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        if self._is_tunnel_cri():
            return (
                ConnectRequestInformation.CRI_TUNNEL_EXT_LENGTH
                if self.individual_address
                else ConnectRequestInformation.CRI_TUNNEL_LENGTH
            )
        return ConnectRequestInformation.CRI_LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        cri_length = raw[0]
        if len(raw) < cri_length:
            raise CouldNotParseKNXIP("CRI data has wrong length")
        if cri_length < ConnectRequestInformation.CRI_LENGTH:
            raise CouldNotParseKNXIP("CRI length too small")
        self.connection_type = ConnectRequestType(raw[1])
        if self._is_tunnel_cri():
            if cri_length == ConnectRequestInformation.CRI_TUNNEL_LENGTH:
                extended = False
            elif cri_length == ConnectRequestInformation.CRI_TUNNEL_EXT_LENGTH:
                extended = True
            else:
                raise CouldNotParseKNXIP("CRI has wrong length")
            self.knx_layer = TunnellingLayer(raw[2])
            self.individual_address = (
                IndividualAddress.from_knx(raw[4:6]) if extended else None
            )
        elif cri_length != ConnectRequestInformation.CRI_LENGTH:
            raise CouldNotParseKNXIP("CRI has wrong length")
        return cri_length

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        _cri = bytes(
            (
                self.calculated_length(),
                self.connection_type.value,
            )
        )
        if self._is_tunnel_cri():
            _cri = _cri + bytes(
                (
                    self.knx_layer.value,
                    0x00,  # Reserved
                )
            )
            if self.individual_address:
                _cri = _cri + self.individual_address.to_knx()

        return _cri

    def __eq__(self, other: object) -> bool:
        """Equal operator."""
        return self.__dict__ == other.__dict__

    def __repr__(self) -> str:
        """Return object as readable string."""
        _tunnel_layer = (
            f'knx_layer="{self.knx_layer.name}" ' if self._is_tunnel_cri() else ""
        )
        _extended = (
            f'individual_address="{self.individual_address}" '
            if self._is_tunnel_cri() and self.individual_address
            else ""
        )
        return (
            "<ConnectRequestInformation "
            f'connection_type="{self.connection_type.name}" '
            f"{_tunnel_layer}"
            f"{_extended}/>"
        )



"""Test Secure Group."""

import asyncio
from unittest.mock import Mock, patch

import pytest

from xknx.cemi import CEMIFrame, CEMILData, CEMIMessageCode
from xknx.exceptions import IPSecureError
from xknx.io.const import DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT, XKNX_SERIAL_NUMBER
from xknx.io.ip_secure import SecureGroup, SecureSequenceTimer
from xknx.knxip import HPAI, KNXIPFrame, RoutingIndication, SecureWrapper, TimerNotify
from xknx.telegram import GroupAddress, Telegram, apci

from ..conftest import EventLoopClockAdvancer

ONE_HOUR_MS = 60 * 60 * 1000


@patch("xknx.io.transport.udp_transport.UDPTransport.connect")
@patch("xknx.io.transport.udp_transport.UDPTransport.send")
class TestSecureGroup:
    """Test class for xknx/io/SecureGroup objects."""

    mock_addr = ("127.0.0.1", 12345)
    mock_backbone_key = bytes.fromhex(
        "0a a2 27 b4 fd 7a 32 31 9b a9 96 0a c0 36 ce 0e"
        "5c 45 07 b5 ae 55 16 1f 10 78 b1 dc fb 3c b6 31"
    )

    def assert_timer_notify(
        self,
        knxipframe: KNXIPFrame,
        timer_value: int | None = None,
        serial_number: bytes | None = None,
        message_tag: bytes | None = None,
        mac: bytes | None = None,
    ) -> bytes:
        """Assert that knxipframe is a TimerNotify."""
        assert isinstance(knxipframe.body, TimerNotify)
        if timer_value is not None:
            assert knxipframe.body.timer_value == timer_value
        if serial_number is not None:
            assert knxipframe.body.serial_number == serial_number
        if message_tag is not None:
            assert knxipframe.body.message_tag == message_tag
        if mac is not None:
            assert knxipframe.body.message_authentication_code == mac
        return knxipframe.body.message_tag

    async def test_no_synchronize(
        self,
        mock_super_send: Mock,
        mock_super_connect: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test synchronisazion not answered."""
        secure_group = SecureGroup(
            local_addr=self.mock_addr,
            remote_addr=(DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT),
            backbone_key=self.mock_backbone_key,
            latency_ms=1000,
        )
        secure_timer = secure_group.secure_timer
        connect_task = asyncio.create_task(secure_group.connect())
        await time_travel(0)
        mock_super_connect.assert_called_once()
        self.assert_timer_notify(
            mock_super_send.call_args[0][0],
            serial_number=XKNX_SERIAL_NUMBER,
        )
        mock_super_send.reset_mock()
        # synchronize timed out
        await time_travel(
            secure_timer.max_delay_time_follower_update_notify
            + 2 * secure_timer.latency_tolerance_ms / 1000
        )
        assert connect_task.done()
        # we are timekeeper so we send TimerNotify after time_keeper_periodic
        assert secure_timer.timekeeper
        assert not secure_timer.sched_update
        await time_travel(secure_timer.MIN_DELAY_TIME_KEEPER_PERIODIC_NOTIFY - 0.01)
        mock_super_send.assert_not_called()
        await time_travel(secure_timer.sync_latency_tolerance_ms / 1000 * 3 + 0.02)
        self.assert_timer_notify(
            mock_super_send.call_args[0][0],
            serial_number=XKNX_SERIAL_NUMBER,
        )
        mock_super_send.reset_mock()
        # test response to invalid timer as timekeeper
        timer_invalid = KNXIPFrame.init_from_body(
            TimerNotify(
                timer_value=0,
                serial_number=bytes.fromhex("00 fa 12 34 56 78"),
                message_tag=bytes.fromhex("12 34"),
                message_authentication_code=bytes.fromhex(
                    "3195051bb981941d57e6c5b55355f341"
                ),
            )
        )
        secure_group.handle_knxipframe(timer_invalid, HPAI(*self.mock_addr))
        assert secure_timer.timekeeper
        assert secure_timer.sched_update
        await time_travel(secure_timer.MIN_DELAY_TIME_KEEPER_UPDATE_NOTIFY - 0.01)
        mock_super_send.assert_not_called()
        await time_travel(secure_timer.sync_latency_tolerance_ms / 1000 * 1 + 0.02)
        self.assert_timer_notify(
            mock_super_send.call_args[0][0],
            message_tag=bytes.fromhex("12 34"),
            serial_number=bytes.fromhex("00 fa 12 34 56 78"),
        )
        # stop
        secure_group.stop()
        assert secure_timer._notify_timer_handle is None

    async def test_synchronize(
        self,
        mock_super_send: Mock,
        mock_super_connect: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test synchronisazion."""
        secure_group = SecureGroup(
            local_addr=self.mock_addr,
            remote_addr=(DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT),
            backbone_key=self.mock_backbone_key,
            latency_ms=1000,
        )
        secure_timer = secure_group.secure_timer
        connect_task = asyncio.create_task(secure_group.connect())
        await time_travel(0)
        mock_super_connect.assert_called_once()
        _message_tag = self.assert_timer_notify(
            mock_super_send.call_args[0][0],
            serial_number=XKNX_SERIAL_NUMBER,
        )
        mock_super_send.reset_mock()
        # incoming
        timer_update = KNXIPFrame.init_from_body(
            TimerNotify(
                timer_value=ONE_HOUR_MS,
                serial_number=XKNX_SERIAL_NUMBER,
                message_tag=_message_tag,
            )
        )
        with patch("xknx.io.ip_secure.SecureSequenceTimer.verify_timer_notify_mac"):
            # TimerNotify MAC is random so we don't verify it in tests
            secure_group.handle_knxipframe(timer_update, HPAI(*self.mock_addr))
        await time_travel(0)
        _leeway_for_ci = 50  # ms
        assert (
            ONE_HOUR_MS
            <= secure_timer.current_timer_value()
            <= ONE_HOUR_MS + _leeway_for_ci
        )
        assert connect_task.done()
        # nothing sent until time_follower_periodic
        assert not secure_timer.timekeeper
        await time_travel(secure_timer.min_delay_time_follower_periodic_notify - 0.01)
        mock_super_send.assert_not_called()
        await time_travel(secure_timer.sync_latency_tolerance_ms / 1000 * 10 + 0.02)
        self.assert_timer_notify(
            mock_super_send.call_args[0][0],
            serial_number=XKNX_SERIAL_NUMBER,
        )
        mock_super_send.reset_mock()
        secure_group.stop()
        assert secure_timer._notify_timer_handle is None

    @patch(  # keep implementation of reschedule
        "xknx.io.ip_secure.SecureSequenceTimer.reschedule",
        side_effect=SecureSequenceTimer.reschedule,
        autospec=True,
    )
    @patch("xknx.io.ip_secure.SecureSequenceTimer._notify_timer_expired")
    @patch("xknx.io.ip_secure.SecureSequenceTimer._monotonic_ms")
    async def test_received_timer_notify(
        self,
        mock_monotonic_ms: Mock,
        _mock_notify_timer_expired: Mock,  # we don't want to actually send here
        mock_reschedule: Mock,
        mock_super_send: Mock,
        mock_super_connect: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test handling of received TimerNotify frames."""
        mock_monotonic_ms.return_value = 0
        secure_group = SecureGroup(
            local_addr=self.mock_addr,
            remote_addr=(DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT),
            backbone_key=self.mock_backbone_key,
            latency_ms=1000,
        )
        secure_timer = secure_group.secure_timer

        # TimerNotify with invalid MAC shall be discarded
        timer_invalid_mac = KNXIPFrame.init_from_body(
            TimerNotify(
                timer_value=ONE_HOUR_MS,
                serial_number=bytes.fromhex("00 fa 12 34 56 78"),
                message_tag=bytes.fromhex("12 34"),
            )
        )
        secure_group.handle_knxipframe(timer_invalid_mac, HPAI(*self.mock_addr))
        mock_reschedule.assert_not_called()
        # E1
        timer_newer = KNXIPFrame.init_from_body(
            TimerNotify(
                timer_value=ONE_HOUR_MS,
                serial_number=bytes.fromhex("00 fa 12 34 56 78"),
                message_tag=bytes.fromhex("12 34"),
                message_authentication_code=bytes.fromhex(
                    "d2fad5657a5788a36cdd8a3ef84c90ab"
                ),
            )
        )
        secure_group.handle_knxipframe(timer_newer, HPAI(*self.mock_addr))
        mock_reschedule.assert_called_once()
        assert secure_timer.current_timer_value() == ONE_HOUR_MS
        assert not secure_timer.timekeeper
        assert not secure_timer.sched_update
        mock_reschedule.reset_mock()
        # E2
        timer_exact = KNXIPFrame.init_from_body(
            TimerNotify(
                timer_value=ONE_HOUR_MS,
                serial_number=bytes.fromhex("00 fa 12 34 56 78"),
                message_tag=bytes.fromhex("12 34"),
                message_authentication_code=bytes.fromhex(
                    "d2fad5657a5788a36cdd8a3ef84c90ab"
                ),
            )
        )
        secure_group.handle_knxipframe(timer_exact, HPAI(*self.mock_addr))
        mock_reschedule.assert_called_once()
        assert secure_timer.current_timer_value() == ONE_HOUR_MS
        assert not secure_timer.timekeeper
        assert not secure_timer.sched_update
        mock_reschedule.reset_mock()

        timer_valid = KNXIPFrame.init_from_body(
            TimerNotify(
                timer_value=ONE_HOUR_MS - secure_timer.sync_latency_tolerance_ms + 1,
                serial_number=bytes.fromhex("00 fa 12 34 56 78"),
                message_tag=bytes.fromhex("12 34"),
                message_authentication_code=bytes.fromhex(
                    "7572b70b4f986d7ae68891a00c0d46d3"
                ),
            )
        )
        secure_group.handle_knxipframe(timer_valid, HPAI(*self.mock_addr))
        mock_reschedule.assert_called_once()
        assert secure_timer.current_timer_value() == ONE_HOUR_MS
        assert not secure_timer.timekeeper
        assert not secure_timer.sched_update
        mock_reschedule.reset_mock()
        # E3
        timer_tolerable_1 = KNXIPFrame.init_from_body(
            TimerNotify(
                timer_value=ONE_HOUR_MS - secure_timer.sync_latency_tolerance_ms,
                serial_number=bytes.fromhex("00 fa 12 34 56 78"),
                message_tag=bytes.fromhex("12 34"),
                message_authentication_code=bytes.fromhex(
                    "3d70cc44607ad9b4a4425de95101a54c"
                ),
            )
        )
        secure_group.handle_knxipframe(timer_tolerable_1, HPAI(*self.mock_addr))
        mock_reschedule.assert_not_called()

        timer_tolerable_2 = KNXIPFrame.init_from_body(
            TimerNotify(
                timer_value=ONE_HOUR_MS - secure_timer.latency_tolerance_ms + 1,
                serial_number=bytes.fromhex("00 fa 12 34 56 78"),
                message_tag=bytes.fromhex("12 34"),
                message_authentication_code=bytes.fromhex(
                    "66b8b2e52196bcce3dce2da7e0dc50ec"
                ),
            )
        )
        secure_group.handle_knxipframe(timer_tolerable_2, HPAI(*self.mock_addr))
        mock_reschedule.assert_not_called()
        # E4
        timer_invalid = KNXIPFrame.init_from_body(
            TimerNotify(
                timer_value=ONE_HOUR_MS - secure_timer.latency_tolerance_ms,
                serial_number=bytes.fromhex("00 fa 12 34 56 78"),
                message_tag=bytes.fromhex("12 34"),
                message_authentication_code=bytes.fromhex(
                    "873d9a5a25b8508446bbd5ff1bb4b465"
                ),
            )
        )
        secure_group.handle_knxipframe(timer_invalid, HPAI(*self.mock_addr))
        mock_reschedule.assert_called_once_with(
            secure_timer,  # self for Mock
            update=(bytes.fromhex("12 34"), bytes.fromhex("00 fa 12 34 56 78")),
        )
        assert not secure_timer.timekeeper
        assert secure_timer.sched_update
        mock_reschedule.reset_mock()
        # E4 from sched_update
        secure_group.handle_knxipframe(timer_invalid, HPAI(*self.mock_addr))
        mock_reschedule.assert_not_called()

    @patch(  # keep implementation of reschedule
        "xknx.io.ip_secure.SecureSequenceTimer.reschedule",
        side_effect=SecureSequenceTimer.reschedule,
        autospec=True,
    )
    @patch("xknx.io.ip_secure.SecureSequenceTimer._notify_timer_expired")
    @patch("xknx.io.ip_secure.SecureSequenceTimer._monotonic_ms")
    @patch("xknx.io.transport.udp_transport.UDPTransport.handle_knxipframe")
    @patch(
        "xknx.io.ip_secure.SecureGroup.decrypt_frame",
        return_value=True,  # we are only interested in timer value
    )
    async def test_received_secure_wrapper(
        self,
        mock_decrypt_frame: Mock,
        mock_super_handle_knxipframe: Mock,
        mock_monotonic_ms: Mock,
        _mock_notify_timer_expired: Mock,  # we don't want to actually send here
        mock_reschedule: Mock,
        mock_super_send: Mock,
        mock_super_connect: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test handling of received SecureWrapper frames."""
        mock_monotonic_ms.return_value = 0
        secure_group = SecureGroup(
            local_addr=self.mock_addr,
            remote_addr=(DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT),
            backbone_key=self.mock_backbone_key,
            latency_ms=1000,
        )
        secure_timer = secure_group.secure_timer
        secure_timer.timer_authenticated = True

        # E5
        wrapper_newer = KNXIPFrame.init_from_body(
            SecureWrapper(
                sequence_information=ONE_HOUR_MS.to_bytes(6, "big"),
            )
        )

        secure_group.handle_knxipframe(wrapper_newer, HPAI(*self.mock_addr))
        mock_super_handle_knxipframe.assert_called_once()
        mock_super_handle_knxipframe.reset_mock()
        mock_reschedule.assert_called_once()
        assert secure_timer.current_timer_value() == ONE_HOUR_MS
        assert not secure_timer.sched_update
        mock_reschedule.reset_mock()
        # E6
        wrapper_exact = KNXIPFrame.init_from_body(
            SecureWrapper(sequence_information=ONE_HOUR_MS.to_bytes(6, "big"))
        )

        secure_group.handle_knxipframe(wrapper_exact, HPAI(*self.mock_addr))
        mock_super_handle_knxipframe.assert_called_once()
        mock_super_handle_knxipframe.reset_mock()
        mock_reschedule.assert_called_once()
        assert secure_timer.current_timer_value() == ONE_HOUR_MS
        assert not secure_timer.timekeeper
        assert not secure_timer.sched_update
        mock_reschedule.reset_mock()

        wrapper_valid = KNXIPFrame.init_from_body(
            SecureWrapper(
                sequence_information=(
                    ONE_HOUR_MS - secure_timer.sync_latency_tolerance_ms + 1
                ).to_bytes(6, "big")
            )
        )

        secure_group.handle_knxipframe(wrapper_valid, HPAI(*self.mock_addr))
        mock_super_handle_knxipframe.assert_called_once()
        mock_super_handle_knxipframe.reset_mock()
        mock_reschedule.assert_called_once()
        assert secure_timer.current_timer_value() == ONE_HOUR_MS
        assert not secure_timer.timekeeper
        assert not secure_timer.sched_update
        mock_reschedule.reset_mock()
        # E7
        wrapper_tolerable_1 = KNXIPFrame.init_from_body(
            SecureWrapper(
                sequence_information=(
                    ONE_HOUR_MS - secure_timer.sync_latency_tolerance_ms
                ).to_bytes(6, "big"),
            )
        )

        secure_group.handle_knxipframe(wrapper_tolerable_1, HPAI(*self.mock_addr))
        mock_super_handle_knxipframe.assert_called_once()
        mock_super_handle_knxipframe.reset_mock()
        mock_reschedule.assert_not_called()

        wrapper_tolerable_2 = KNXIPFrame.init_from_body(
            SecureWrapper(
                sequence_information=(
                    ONE_HOUR_MS - secure_timer.latency_tolerance_ms + 1
                ).to_bytes(6, "big"),
                serial_number=bytes.fromhex("00 fa 12 34 56 78"),
                message_tag=bytes.fromhex("12 34"),
            )
        )

        secure_group.handle_knxipframe(wrapper_tolerable_2, HPAI(*self.mock_addr))
        mock_super_handle_knxipframe.assert_called_once()
        mock_super_handle_knxipframe.reset_mock()
        mock_reschedule.assert_not_called()
        # E8
        wrapper_invalid = KNXIPFrame.init_from_body(
            SecureWrapper(
                sequence_information=(
                    ONE_HOUR_MS - secure_timer.latency_tolerance_ms
                ).to_bytes(6, "big"),
                serial_number=bytes.fromhex("00 fa 12 34 56 78"),
                message_tag=bytes.fromhex("12 34"),
            )
        )

        secure_group.handle_knxipframe(wrapper_invalid, HPAI(*self.mock_addr))
        mock_reschedule.assert_called_once_with(
            secure_timer,  # self for Mock
            update=(bytes.fromhex("12 34"), bytes.fromhex("00 fa 12 34 56 78")),
        )
        assert not secure_timer.timekeeper
        assert secure_timer.sched_update
        mock_reschedule.reset_mock()
        # E8 from sched_update
        secure_group.handle_knxipframe(wrapper_invalid, HPAI(*self.mock_addr))
        mock_super_handle_knxipframe.assert_not_called()
        mock_reschedule.assert_not_called()

    @patch(  # keep implementation of reschedule
        "xknx.io.ip_secure.SecureSequenceTimer.reschedule",
        side_effect=SecureSequenceTimer.reschedule,
        autospec=True,
    )
    @patch("xknx.io.ip_secure.SecureSequenceTimer._notify_timer_expired")
    @patch("xknx.io.ip_secure.SecureSequenceTimer._monotonic_ms")
    async def test_send_secure_wrapper(
        self,
        mock_monotonic_ms: Mock,
        _mock_notify_timer_expired: Mock,  # we don't want to actually send here
        mock_reschedule: Mock,
        mock_super_send: Mock,
        mock_super_connect: Mock,
        time_travel: EventLoopClockAdvancer,
    ) -> None:
        """Test sending SecureWrapper frames."""
        mock_monotonic_ms.return_value = 0
        secure_group = SecureGroup(
            local_addr=self.mock_addr,
            remote_addr=(DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT),
            backbone_key=self.mock_backbone_key,
            latency_ms=1000,
        )
        secure_timer = secure_group.secure_timer
        secure_timer.timer_authenticated = True

        raw_test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(
                Telegram(
                    destination_address=GroupAddress("1/2/3"),
                    payload=apci.GroupValueRead(),
                )
            ),
        ).to_knx()

        assert not secure_timer.sched_update
        secure_group.send(
            KNXIPFrame.init_from_body(RoutingIndication(raw_cemi=raw_test_cemi))
        )
        mock_reschedule.assert_called_once()
        mock_reschedule.reset_mock()
        # no reschedule when sched_update is true
        secure_timer.reschedule(update=(bytes(2), bytes(6)))
        assert secure_timer.sched_update
        mock_reschedule.reset_mock()
        secure_group.send(
            KNXIPFrame.init_from_body(RoutingIndication(raw_cemi=raw_test_cemi))
        )
        mock_reschedule.assert_not_called()

    @patch("xknx.io.ip_secure.SecureSequenceTimer._notify_timer_expired")
    @patch("xknx.io.ip_secure.SecureSequenceTimer._monotonic_ms")
    async def test_send_secure_wrapper_timer_overflow(
        self,
        mock_monotonic_ms: Mock,
        _mock_notify_timer_expired: Mock,  # we don't want to actually send here
        mock_super_send: Mock,
        mock_super_connect: Mock,
    ) -> None:
        """Test raising when secure timer overflows."""
        mock_monotonic_ms.return_value = 0xFF_FF_FF_FF_FF_FF + 1
        secure_group = SecureGroup(
            local_addr=self.mock_addr,
            remote_addr=(DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT),
            backbone_key=self.mock_backbone_key,
            latency_ms=1000,
        )
        secure_timer = secure_group.secure_timer
        secure_timer.timer_authenticated = True

        raw_test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(
                Telegram(
                    destination_address=GroupAddress("1/2/3"),
                    payload=apci.GroupValueRead(),
                )
            ),
        ).to_knx()

        with pytest.raises(IPSecureError):
            secure_group.send(
                KNXIPFrame.init_from_body(RoutingIndication(raw_cemi=raw_test_cemi))
            )

    async def test_receive_plain_frames(
        self,
        mock_super_send: Mock,
        mock_super_connect: Mock,
    ) -> None:
        """Test class for KNXnet/IP secure routing."""
        frame_received_mock = Mock()
        secure_group = SecureGroup(
            local_addr=self.mock_addr,
            remote_addr=(DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT),
            backbone_key=self.mock_backbone_key,
            latency_ms=1000,
        )
        secure_group.register_callback(frame_received_mock)
        plain_routing_indication = bytes.fromhex("0610 0530 0010 2900b06010fa10ff0080")
        secure_group.data_received_callback(
            plain_routing_indication, ("192.168.1.2", 3671)
        )
        frame_received_mock.assert_not_called()
        plain_search_response = bytes.fromhex(
            "0610020c006608010a0100280e57360102001000000000082d40834de000170c"
            "000ab3274a3247697261204b4e582f49502d526f757465720000000000000000"
            "000000000e02020203020402050207010901140700dc10f1fffe10f2ffff10f3"
            "ffff10f4ffff"
        )
        secure_group.data_received_callback(
            plain_search_response, ("192.168.1.2", 3671)
        )
        frame_received_mock.assert_called_once()
